/// @file GLWindow.cpp
/// @brief basic implementation file for the GLWindow class
#include "GLWindow.h"
#include <iostream>
#include "ngl/Camera.h"
#include "ngl/Colour.h"
#include "ngl/Light.h"
#include "ngl/Matrix.h"
#include "ngl/Transformation.h"
#include "ngl/TransformStack.h"
#include "ngl/Material.h"
#include "ngl/NGLInit.h"
#include "ngl/Obj.h"
#include "ngl/Util.h"
#include "ngl/VBOPrimitives.h"
#include "ngl/ShaderManager.h"

#include <QDebug>

const  std::string GLWindow::s_vboNames[13]=
{
  "sphere",
  "cylinder",
  "cone",
  "disk",
  "plane",
  "torus",
  "teapot",
  "octahedron",
  "dodecahedron",
  "icosahedron",
  "tetrahedron",
  "football",
  "cube"

};



//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{

  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // Now set the initial GLWindow attributes to default values
  // Roate is false
  m_rotate=false;
  // mouse rotation values set to 0
  m_spinXFace=0;
  m_spinYFace=0;
  m_drawIndex=6;
  m_drawNormals=false;
  m_drawVerts=false;
  /// set all our matrices to the identity
  m_transform=1.0;
  m_rotate=1.0;
  m_translate=1.0;
  m_scale=1.0;
  m_normalSize=6.0;
  m_colour.set(0.5,0.5,0.5);
  m_material.setDiffuse(m_colour);
  m_material.setSpecular(ngl::Colour(0.2,0.2,0.2));
  m_material.setAmbient(ngl::Colour());
  m_material.setSpecularExponent(20);
  m_material.setRoughness(0);
  m_material.setEmisive(ngl::Colour());
  m_matrixOrder=GLWindow::RTS;
  m_euler=1.0;
}

// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{

  glClearColor(0.4f, 0.4f, 0.4f, 1.0f);			   // Grey Background
  // enable depth testing for drawing
  glEnable(GL_DEPTH_TEST);
  // Now we will create a basic Camera from the graphics library
  // This is a static camera so it only needs to be set once
  // First create Values for the camera position
  ngl::Vector From(0,0,8);
  ngl::Vector To(0,0,0);
  ngl::Vector Up(0,1,0);
  ngl::NGLInit *Init = ngl::NGLInit::instance();
  #ifdef WIN32
  	initGlew();
  #endif
  Init->initGlew();

  m_cam= new ngl::Camera(From,To,Up,ngl::PERSPECTIVE);
  // set the shape using FOV 45 Aspect Ratio based on Width and Height
  // The final two are near and far clipping planes of 0.5 and 10
  m_cam->setShape(100,(float)720.0/576.0,0.5,10,ngl::PERSPECTIVE);
  // now to load the shader and set the values
  // grab an instance of shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // load a frag and vert shaders
  shader->loadShader("Blinn","shaders/Vertex.vs","shaders/Fragment.fs");
  // set this as the active shader
  shader->useShader("Blinn");
  // now pass the modelView and projection values to the shader
  shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cam->getModelView());
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
  m_light = new ngl::Light(ngl::Vector(1,1,1,1),ngl::Colour(1,1,1,1),ngl::LIGHTREMOTE);

  // load a frag and vert shaders for colour no shading
  shader->loadShader("Colour","shaders/Colour.vs","shaders/Colour.fs");
  // set this as the active shader
  shader->useShader("Colour");
  shader->setShaderParamFromMatrix("Colour","ViewMatrix",m_cam->getModelView());
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());
  m_light->enable();

  /// now create our primitives for drawing later

  ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
  prim->createVBOSphere("sphere",1.0,40);
  prim->createVBOCylinder("cylinder",0.5,1.4,40,40);
  prim->createVBOCone("cone",0.5,1.4,20,20);
  prim->createVBODisk("disk",0.5,40);
  prim->createVBOQuadPlane("plane",1,1,10,10,ngl::Vector(0,1,0));
  prim->createVBOTorus("torus",0.15,0.4,40,40);
  // set the bg colour
  glClearColor(0.5,0.5,0.5,0.0);
  m_axis = new Axis("Blinn",2.5);


}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  qDebug()<<"resize event"<<_w<<" "<<_h;
  glViewport(0,0,_w,_h);
  m_cam->setShape(100,(float)_w/_h,0.5,10,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
  shader->setShaderParamFromMatrix("Colour","projectionMatrix",m_cam->getProjection());

}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
void GLWindow::paintGL()
{
  // grab an instance of the shader manager
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  // Rotation based on the mouse position for our global
  // transform

  m_light->enable();

  m_transform.identity();

  if (m_matrixOrder == GLWindow::RTS)
  {
    m_transform=m_rotate*m_translate*m_scale;
  }

  else if (m_matrixOrder == GLWindow::TRS)
  {
    m_transform=m_translate*m_rotate*m_scale;
  }
  else if (m_matrixOrder == GLWindow::EULERTS)
  {
    m_transform=m_translate*m_euler*m_scale;
  }
  else if (m_matrixOrder == GLWindow::TEULERS)
  {
    m_transform=m_euler*m_translate*m_scale;
  }
  else if (m_matrixOrder == GLWindow::TEULERS)
  {
    m_transform=m_euler*m_translate*m_scale;
  }

  else if (m_matrixOrder == GLWindow::GIMBALLOCK )
  {
    m_transform=m_translate*m_gimbal*m_scale;
  }

  emit matrixDirty(m_transform);
  // now set this value in the shader for the current ModelMatrix
  m_material.use();

  shader->useShader("Blinn");

  ngl::TransformStack txStack;
  ngl::Transformation tx;
  txStack.pushTransform();
  {

    // set the global rotation
    tx.setRotation(m_spinXFace,m_spinYFace,0.0);
    txStack.setCurrent(tx);

    txStack.pushTransform();
    {
      txStack.setCurrent(m_transform);
      txStack.loadGlobalMatrixToShader("Blinn","ModelMatrix");

      // get the VBO instance and draw the built in teapot
      ngl::VBOPrimitives *prim=ngl::VBOPrimitives::instance();
      prim->draw( s_vboNames[m_drawIndex]);

      shader->useShader("Colour");
      txStack.loadGlobalMatrixToShader("Colour","ModelMatrix");
      if(m_drawNormals)
      {
        shader->setShaderParam4f("Colour","Colour",1.0,1.0,0.0,1.0);
        prim->drawNormals(s_vboNames[m_drawIndex],float(m_normalSize));
      }
      if(m_drawVerts)
      {
        shader->setShaderParam4f("Colour","Colour",1.0,1.0,1.0,1.0);
        prim->drawVerts(s_vboNames[m_drawIndex]);
      }
   }
  txStack.popTransform();
  }
  m_axis->draw( m_spinXFace,m_spinYFace);

txStack.popTransform();


}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseMoveEvent (
                               QMouseEvent * _event
                              )
{
  // note the method buttons() is the button state when event was called
  // this is different from button() which is used to check which button was
  // pressed when the mousePress/Release event is generated
  if(m_rotateMouse && _event->buttons() == Qt::LeftButton)
  {
    m_spinYFace = ( m_spinYFace + (_event->x() - m_origX) ) % 360 ;
    m_spinXFace = ( m_spinXFace + (_event->y() - m_origY) ) % 360 ;
    m_origX = _event->x();
    m_origY = _event->y();
  }
  // re-draw GL
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mousePressEvent (
                                QMouseEvent * _event
                               )
{
  // this method is called when the mouse button is pressed in this case we
  // store the value where the maouse was clicked (x,y) and set the Rotate flag to true
  if(_event->button() == Qt::LeftButton)
  {
    m_origX = _event->x();
    m_origY = _event->y();
    m_rotateMouse =true;
  }
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::mouseReleaseEvent (
                                  QMouseEvent * _event
                                 )
{
  // this event is called when the mouse button is released
  // we then set Rotate to false
  if (_event->button() == Qt::LeftButton)
  {
    m_rotateMouse=false;
  }
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::vboChanged(
                          int _index
                         )
{
  m_drawIndex=_index;
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::toggleNormals(
                              bool _value
                            )
{
  m_drawNormals=_value;
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::toggleVerts(
                            bool _value
                          )
{
  m_drawVerts=_value;
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::toggleWireframe(
                                bool _value
                              )
{
   if( _value == true)
   {
     glPolygonMode(GL_FRONT_AND_BACK,GL_LINE);
   }
   else
   {
     glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
   }
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setNormalSize(
                              int _value
                            )
{
  m_normalSize=_value;
  updateGL();
}
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setScale(
                        float _x,
                        float _y,
                        float _z
                        )

{
  m_scale.scale(_x,_y,_z);
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setTranslate(
                            float _x,
                            float _y,
                            float _z
                            )
{

  m_translate.translate(_x,_y,_z);
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setRotate(
                          float _x,
                          float _y,
                          float _z
                          )
{
  ngl::Matrix rx=1.0;
  ngl::Matrix ry=1.0;
  ngl::Matrix rz=1.0;
  rx.rotateX(_x);
  ry.rotateY(_y);
  rz.rotateZ(_z);
  m_rotate=rz*ry*rx;
  // now for the incorrect gimbal 1
  m_gimbal.identity();
  m_gimbal.rotateX(_x);
  m_gimbal.rotateY(_y);
  m_gimbal.rotateZ(_z);

  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setColour(
                          float _r,
                          float _g,
                          float _b
                        )
{
  m_colour.set(_r,_g,_b);
  m_material.setDiffuse(m_colour);
  m_material.use();
  updateGL();
}


//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setMatrixOrder(
                              int _index
                              )
{
  switch(_index)
  {
    case 0 : { m_matrixOrder=GLWindow::RTS;   break; }
    case 1 : { m_matrixOrder=GLWindow::TRS;   break; }
    case 2 : { m_matrixOrder=GLWindow::GIMBALLOCK; break; }
    case 3 : { m_matrixOrder=GLWindow::EULERTS; break; }
    case 4 : { m_matrixOrder=GLWindow::TEULERS; break; }
    default : break;
  }
updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::setEuler(
                        float _angle,
                        float _x,
                        float _y,
                        float _z
                       )
{
  m_euler=1.0;
  m_euler.euler(_angle,_x,_y,_z);
  updateGL();
}

//----------------------------------------------------------------------------------------------------------------------
void GLWindow::resetMouse()
{
m_spinXFace=0;
m_spinYFace=0;
m_origX=0;
m_origY=0;
updateGL();
}
//----------------------------------------------------------------------------------------------------------------------
